# SIS as a compartmented model
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *

import random

class SIS(CompartmentedModel):
    '''The Susceptible-Infected-Susceptible :term:`compartmented model of disease`.
    Susceptible nodes are infected by infected neighbours, and recover back
    to the susceptible state (which allows future re-infection, unlike for
    :class:`SIR`.'''

    # the model parameters
    P_INFECTED = 'pInfected'  #: Parameter for probability of initially being infected.
    P_INFECT = 'pInfect'      #: Parameter for probability of infection on contact.
    P_RECOVER = 'pRecover'    #: Parameter for probability of recovery.
    
    # the possible dynamics states of a node for SIR dynamics
    SUSCEPTIBLE = 'S'         #: Compartment for nodes susceptible to infection.
    INFECTED = 'I'            #: Compartment for nodes infected.

    # the edges at which dynamics can occur
    SI = 'SI'                 #: Edge able to transmit infection.

    def __init__( self ):
        super(SIS, self).__init__()

    def build( self, params ):
        '''Build the SIS model.

        :param params: the model parameters'''
        pInfected = params[self.P_INFECTED]
        pInfect = params[self.P_INFECT]
        pRecover = params[self.P_RECOVER]

        self.addCompartment(self.SUSCEPTIBLE, 1 - pInfected)
        self.addCompartment(self.INFECTED, pInfected)

        self.addLocus(self.SUSCEPTIBLE, self.INFECTED, name = self.SI)
        self.addLocus(self.INFECTED)

        self.addEvent(self.INFECTED, pRecover, lambda t, g, e: self.recover(g, e))
        self.addEvent(self.SI, pInfect, lambda t, g, e: self.infect(g, e))

    def infect( self, g, (n, m) ):
        self.changeCompartment(g, n, self.INFECTED)
        self.markOccupied(g, (n, m))

    def recover( self, g, n ):
        self.changeCompartment(g, n, self.SUSCEPTIBLE)
    
                
   



# Loci for dynamics in compartmented models model
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *

import numpy
import random

class Locus(object):
    '''The locus of dynamics. A locus is where dynamics happens, allowing the
    compartments of nodes to be changed and other effects to be coded-up. Loci
    are filled with nodes or edges, typically populated and re-populated as the
    dynamics moves nodes between components.'''

    def __init__( self, name ):
        '''Create a new named locus.

        :param name: the locus; name'''
        super(Locus, self).__init__()
        self._name = name
        self._elements = set()

    def name( self ):
        '''Returns the name of the locus.

        :returns: the locus' name'''
        return self_name
        
    def __len__( self ):
        '''Return the number of elements at the locus.

        :returns: the number of elements'''
        return len(self._elements)

    def elements( self ):
        '''Return the underlying elements of the locus.

        :returns: the elements'''
        return self._elements
    
    def draw( self ):
        '''Draw a random element from the locus.

        :returns: a random element at the locus'''
        e = (random.sample(self._elements, 1))[0]
        #self._elements.remove(e)
        return e

    def leaveHandler( self, m, g, n, c ):
        '''Handler for when a node leaves a compartment., Must be overridden
        by sub-classes.

        :param m: the model
        :param g: the network
        :param n: the node
        :param c: the compartment the node is leaving'''
        raise NotYetImplemented('leaveHandler()')

    def enterHandler( self, m, g, n, c ):
        '''Handler for when a node enters a compartment., Must be overridden
        by sub-classes.

        :param m: the model
        :param g: the network
        :param n: the node
        :param c: the compartment the node is entering'''
        raise NotYetImplemented('enterHandler()')


class NodeLocus(Locus):
    '''A locus for dynamics occurring at a single node. Node loci
    contain nodes, typically all in a single compartment.'''

    def __init__( self, name, c ):
        '''Create a locus for nodes in the given compartment.

        :param name: the locus' name
        :param c: the compartment''' 
        super(NodeLocus, self).__init__(name)
        self._compartment = c
        
    def leaveHandler( self, m, g, n, c ):
        '''Node leaves the right compartment, remove it from the locus

        :param m: the model
        :param g: the network
        :param n: the node
        :param c: the compartment the node is leavinging'''
        #print 'node {n} leaves {c}'.format(n = n, c = self._compartment)
        self._elements.remove(n)

    def enterHandler( self, m, g, n, c ):
        '''Node enters the right compartment, add it to the locus

        :param m: the model
        :param g: the network
        :param n: the node
        :param c: the compartment the node is entering'''
        #print 'node {n} enters {c}'.format(n = n, c = self._compartment)
        self._elements.add(n)


class EdgeLocus(Locus):
    '''A locus for dynamics occurring at an edge. Edge loci contain
    edges, typically with the endpoint nodes in different compartments. The
    edges within a locus change as nodes move between compartments.'''

    def __init__( self, name, l, r ):
        '''Create a locus for an edge with endpoints in the
        given compartments. Edges are treated as directed, in the
        sense that the edge will always be manipulated according to
        the given orientation.

        :param name: the locus' name
        :param l: the left compartment 
        :param r: the right compartment''' 
        super(EdgeLocus, self).__init__(name)
        self._left = l
        self._right = r

    def leaveHandler( self, m, g, n, c ):
        '''Node leaves one of the edge's compartments, remove any incident edges
        that no longer have the correct orientation.

        :param m: the model
        :param g: the network
        :param n: the node
        :param c: the compartment the node is leaving'''
        for (nn, mm) in g.edges_iter(n):
            if (g.node[nn][m.COMPARTMENT] == self._right) and (g.node[mm][m.COMPARTMENT] == self._left):
                #print 'edge ({m}, {n}) leaves {l}'.format(n = nn, m = mm, l = self._name)
                self._elements.remove((mm, nn))
            else:
                if (g.node[nn][m.COMPARTMENT] == self._left) and (g.node[mm][m.COMPARTMENT] == self._right):
                    #print 'edge ({n}, {m}) leaves {l}'.format(n = nn, m = mm, l = self._name)
                    self._elements.remove((nn, mm))

    def enterHandler( self, m, g, n, c ):
        '''Node enters one of the edge's compartments, add any incident edges
        that now have the correct orientation.

        :param m: the model
        :param g: the network
        :param n: the node
        :param c: the compartment the node is entering'''
        for (nn, mm) in g.edges_iter(n):
            if (g.node[nn][m.COMPARTMENT] == self._right) and (g.node[mm][m.COMPARTMENT] == self._left):
                #print 'edge ({m}, {n}) enters {l}'.format(n = nn, m = mm, l = self._name)
                self._elements.add((mm, nn))
            else:
                if (g.node[nn][m.COMPARTMENT] == self._left) and (g.node[mm][m.COMPARTMENT] == self._right):
                    #print 'edge ({n}, {m}) enters {l}'.format(n = nn, m = mm, l = self._name)
                    self._elements.add((nn, mm))



# Gillespie dynamics base class
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *

import epyc
import math
import numpy

class StochasticDynamics(Dynamics):
    '''A dynamics that runs stochastically in :term:`continuous time`. This is a
    very efficient and statistically exact approach, but requires that the
    statistical properties of the events making up the process are known.'''
    
    def __init__( self, g = None ):
        '''Create a dynamics, optionally initialised to run on the given network.
        
        :param g: prototype network to run the dynamics over (optional)'''
        super(StochasticDynamics, self).__init__(g)

    def eventRateDistribution( self, t ):
        '''Return the event distribution, a sequence of (r, f) pairs
        where r is the rate at which an event happens and
        f is the event function called to make it happen. This method
        must be overridden in sub-classes. Note that it's a rate we want,
        not a probability: the former can be obtained from the latter
        simply by multiplying the event probability by the number of
        times it's possible in the current network, which is often a
        population of nodes or edges in a given state.
        
        It is perfectly fine for an event to have a zero rate. The process
        is assumed to have reached equilibrium if all events have zero rates.

        :param t: current time
        :returns: the event rate distribution'''
        raise NotYetImplementedError('eventRateDistribution()')

    def do( self, params ):
        '''Run the simulation using Gillespie dynamics. The process terminates
        when either there are no events with zero rates or when :meth:`at_equilibrium`
        return True.

        :param params: the experimental parameters
        :returns: the experimental results dict'''
        
        # run the dynamics
        g = self.network()
        t = 0
        events = 0
        while not self.at_equilibrium(t):
            # pull the transition dynamics at this timestep
            transitions = self.eventRateDistribution(t)
            
            # compute the total rate of transitions for the entire network
            a = 0.0
            for (_, r, _) in transitions:
                a = a + r
            if a == 0:       
                break              # no events with non-zero rates 
            
            # calculate the timestep delta
            r1 = numpy.random.random()
            dt = (1.0 / a) * math.log(1.0 / r1)

            # advance time by the timestep
            t = t + dt
            
            # calculate which transition happens
            if len(transitions) == 1:
                # if there's only one, that's the one that happens
                (l, _, f) = transitions[0]
            else:
                # otherwise, choose one at random based on the rates
                r2 = numpy.random.random()
                xc = r2 * a
                k = 0
                (l, xs, f) = transitions[k]
                while xs < xc:
                    k = k + 1
                    (l, xsp, f) = transitions[k]
                    xs = xs + xsp

            # draw a random element from the chosen locus
            e = l.draw()
            
            # perform the event by calling the event function
            f(t, g, e)
            
            # increment the event counter
            events = events + 1

        # add some more metadata
        (self.metadata())[self.TIME] = t
        (self.metadata())[self.EVENTS] = events

        # report results
        rc = self.experimentalResults()
        return rc




# Networks dynamics simulation base class
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

import epyc
import networkx


class Dynamics(epyc.Experiment, object):
    '''A dynamical process over a network. This is the abstract base class
    for implementing different kinds of dynamics as computational experiments
    suitable for running under. Sub-classes provide synchronous and stochastic
    (Gillespie) simulation dynamics.'''

    # Additional metadata elements
    TIME = 'simulation_time'      #: Metadata element holding the logical simulation end-time.
    EVENTS = 'simulation_events'  #: Metadata element holding the number of events that happened.

    # the default maximum simulation time
    DEFAULT_MAX_TIME = 20000      #: Default maximum simulation time.
    
    def __init__( self, g = None ):
        '''Create a dynamics, optionally initialised to run on the given network.
        The network (if provided) is treated as a prototype that is copied before
        each individual simulation experiment.
        
        :param g: prototype network (optional)'''
        super(Dynamics, self).__init__()
        self._graphPrototype = g
        self._graph = None
        self._maxTime = self.DEFAULT_MAX_TIME

    def network( self ):
        '''Return the network this dynamics is running over.

        :returns: the network'''
        return self._graph

    def setNetworkPrototype( self, g ):
        '''Set the network the dynamics will run over. This will be
        copied for each run of an individual experiment.

        :param g: the network'''
        self._graphPrototype = g

    def setMaximumTime( self, t ):
        '''Set the maximum default simulation time. The default is given
        by :attr:`DEFAULT_MAX_TIME`.

        param: t: the maximum time'''
        self._maxTime = t
        
    def at_equilibrium( self, t ):
        '''Test whether the model is an equilibrium. Override this method to provide
        alternative and/or faster simulations.
        
        :param t: the current simulation timestep
        :returns: True if we're done'''
        return (t >= self._maxTime)

    def setUp( self, params ): 
        '''Before each experiment, create a working copy of the prototype network.

        :param params: parameters of the experiment'''

        # perform the default setup
        super(Dynamics, self).setUp(params)

        # make a copyt of the network prototype
        self._graph = self._graphPrototype.copy()

    def tearDown( self ):
        '''At the end of each experiment, throw away the copy.'''

        # perform the default tear-down
        super(Dynamics, self).tearDown()

        # throw away the worked-on model
        self._graph = None
        
    



# Synchronous dynamics base class
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *

import epyc
import networkx
import numpy

class SynchronousDynamics(Dynamics):
    '''A dynamics that runs synchronously in discrete time, applying local
    rules to each node in the network. These are simple to understand and
    simple to code for many cases, but can be statistically inexact and slow
    for large systems.'''

    # additional metadata
    TIMESTEPS_WITH_EVENTS = 'timesteps_with_events'  #: Metadata element holding the number timesteps that actually had events occur within them

    def __init__( self, g = None ):
        '''Create a dynamics, optionally initialised to run on the given prototype
        network.
        
        :param g: prototype network to run over (optional)'''
        super(SynchronousDynamics, self).__init__(g)

    def dynamics( self, t, params ):
        '''Run one step of the discrete-time dynamics. This method must
        be overridden by sub-classes.

        :param t: the current simulation time
        :param params: the simulation parameters
        :returns: the number of events that happened in the timestep'''
        raise NotYetImplemented('dynamics()')
    
    def do( self, params ):
        '''Synchronous dynamics. We apply :meth:`dynamics` at each timestep
        and then check for completion using :meth:`at_equilibrium`.
        
        :param params: the parameters of the simulation
        :returns: a dict of experimental results'''

        # run the dynamics
        t = 0
        events = 0
        timestepEvents = 0
        while not self.at_equilibrium(t):
            # run a step
            nev = self.dynamics(t, params)
            if nev > 0:
                events = events + nev
                timestepEvents = timestepEvents + 1

            # advance to the next timestep
            t = t + 1

        # add some more metadata
        (self.metadata())[self.TIME] = t
        (self.metadata())[self.EVENTS] = events
        (self.metadata())[self.TIMESTEPS_WITH_EVENTS] = timestepEvents

        # report results
        rc = self.experimentalResults()
        return rc



# Test suite
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

import unittest
from .compartmentedmodel import CompartmentedModelTest
from .sir import SIRTest
from .sis import SISTest

cmsuite = unittest.TestLoader().loadTestsFromTestCase(CompartmentedModelTest)
sirsuite = unittest.TestLoader().loadTestsFromTestCase(SIRTest)
sissuite = unittest.TestLoader().loadTestsFromTestCase(SISTest)

suite = unittest.TestSuite([ cmsuite,
                             sirsuite,
                             sissuite ])

if __name__ == '__main__':
    unittest.TextTestRunner(verbosity = 2).run(suite)



# Setup for epydemic
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from setuptools import setup

with open('README.rst') as f:
    longDescription = f.read()

setup(name = 'epydemic',
      version = '0.2.0',
      description = 'Epidemic network simulations in Python',
      long_description = longDescription,
      url = 'http://github.com/simoninireland/epydemic',
      author = 'Simon Dobson',
      author_email = 'simon.dobson@computer.org',
      license = 'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
      classifiers = [ 'Development Status :: 4 - Beta',
                      'Intended Audience :: Science/Research',
                      'Intended Audience :: Developers',
                      'Programming Language :: Python :: 2.7',
                      'Topic :: Scientific/Engineering' ],
      packages = [ 'epydemic' ],
      zip_safe = True,
      install_requires = [ "numpy", "networkx", "epyc", ])



# Synchronous dynamics base class
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *

import epyc
import networkx
import numpy
from copy import copy

class CompartmentedSynchronousDynamics(SynchronousDynamics):
    '''A :term:`synchronous dynamics` running a compartmented model. The
    behaviour of the simulation is completely described within the model
    rather than here.'''
        
    def __init__( self, m, g = None ):
        '''Create a dynamicsover the given disease model, optionally
        initialised to run on the given prototype network.
        
        :param m: the model
        :param g: prototype network to run over (optional)'''
        super(CompartmentedSynchronousDynamics, self).__init__(g)
        self._model = m

    def setUp( self, params ):
        '''Set up the experiment for a run. This performs the default action
        of copying the prototype network and then builds the model and
        uses it to initialise the nodes into the various compartments
        according to the parameters.

        :params params: the experimental parameters'''
        
        # perform the default setup
        super(CompartmentedSynchronousDynamics, self).setUp(params)

        # build the model
        self._model.build(params)

        # initialise the network from the model
        g = self.network()
        self._model.setUp(g, params)

    def dynamics( self, t, params ):
        '''Run a single step of the model over the network.
        
        :param t: the current timestep
        :param params: the parameters of the simulation
        :returns: the number of dynamic events that happened in this timestep'''
        g = self.network()
        events = 0
        
        # retrieve all the events, their loci, probabilities, and event functions
        dist = self._model.eventDistribution(t)

        # run through all the events
        for (l, p, f) in dist:
            # run through every possible element on which this event may occur
            for e in copy(l.elements()):
                # test for occurrance of the event on this element
                if numpy.random.random() < p:
                    # yes, perform the event
                    f(t, g, e)

                    # update the event count
                    events = events + 1

        return events

    def experimentalResults( self ):
        '''Report the model's experimental results.

        :returns: the results as seen by the model'''
        return self._model.results(self.network())
